#include "classify.h"

#include <time.h>
#include <stdio.h>
#include <omp.h>

#include <tr1/unordered_map>
#include <iomanip>
#include <limits>
#include <iostream>
#include <iterator>
#include <algorithm>
#include <string>

#define min(x,y) ((x)<(y)?(x):(y))
#define max(x,y) ((x)>(y)?(x):(y))
#define abs(x)   ((x) >0 ?(x):-(x))

#define MAX_CHAR_PER_LINE 200000

#define DEBUG 0
#define DOUBLE_MAX_VALUE std::numeric_limits<double>::infinity()

// the best window-sizes to choose from
double factor = 0.95;

std::vector< TimeSeries* > trainSamples;
std::vector< TimeSeries* > testSamples;

std::string datasets[] = {
		"Two_Patterns",
		"ChlorineConcentration",
		"wafer",
		"MedicalImages",
		"FaceAll",
		"OSULeaf",
		"Adiac",
		"SwedishLeaf",
		"yoga",
		"fish",
		"Lighting7",
		"Lighting2",
		"Trace",
		"synthetic_control",
		"FacesUCR",
		"CinC_ECG_torso",
		"MALLAT",
		"Symbols",
		"Coffee",
		"ECG200",
		"FaceFour",
		"OliveOil",
		"Gun_Point",
		"Beef",
		"DiatomSizeReduction",
		"CBF",
		"ECGFiveDays",
		"TwoLeadECG",
		"SonyAIBORobot SurfaceII",
		"MoteStrain",
		"ItalyPowerDemand",
		"SonyAIBORobot Surface",
//		"StarLightCurves100",
//		"StarLightCurves250",
//		"StarLightCurves500",
//		"StarLightCurves",
};

void loadDatset(const char* filename, std::vector< TimeSeries* > & samples) {

	FILE* f;
	f = fopen(filename, "r");
	if (f==NULL) {
		std::cout << "Error!! Input file " << filename << " not found." << std::endl;
		exit(0);
	}

	char buf[MAX_CHAR_PER_LINE];
	char* tmp;
	double classLabel = 0;

	while (fgets(buf, MAX_CHAR_PER_LINE, f)) {
		std::vector<double> tsData;

		// split buf into tokens
		tmp = strtok(buf, " \r\n");

		// read label
		classLabel = atof(tmp);

		// read data
		tmp = strtok(NULL, " \r\n");
		while(tmp != NULL) {
			tsData.push_back(atof(tmp));
			tmp = strtok(NULL," \r\n");
		}

		TimeSeries* ts = new TimeSeries(tsData, classLabel);
		ts->norm(true);
		samples.push_back(ts);

	}

	std::cout << "Done reading: " << filename
			<< ". " << samples.size()
			<< " time series of length "
			<< samples[0]->getSize() << std::endl;

	fclose(f);
}

void classify() {
	double startTime = omp_get_wtime();

	int correct = predictEuclideanDistance(testSamples, trainSamples);
	outputResult(correct, startTime, testSamples.size());

	// training
	int minWindowLength = 5;
	int maxWindowLength = trainSamples[0]->getSize();
	bool normMean = false;

	int bestWindowLength = 0;
	bool bestNormedMean = false;
	int bestCorrectTesting = 0;
	int bestCorrectTraining = 0;

	bool normed[]{true, false};
	for (auto norm : normed) {
		startTime = omp_get_wtime();

		int currentCorrect = 0;
		int currentWindowLength = 0;
		const std::vector<WindowSizes> windows = fit(
				maxWindowLength, minWindowLength, currentCorrect, currentWindowLength, norm);

		std::cout << "Training ShotGun:\t"
				<< "best window-size:\t" << currentWindowLength
				<< "\t normed: \t"
				<< (norm? "true":"false") << std::endl;
		outputResult(currentCorrect, startTime, trainSamples.size());

		std::cout << "Testing ShotGun:\t"
				<< "best window-size:\t" << currentWindowLength
				<< "\t normed: \t"
				<< (norm? "true":"false") << std::endl;


		// normal shotgun classifier
//		int correctTesting = predict(currentWindowLength, 0, norm, testSamples, trainSamples).correct;

		// shotgun ensemble classifier
		int correctTesting = predictEnsemble(windows, currentCorrect, norm, testSamples, trainSamples).correct;
		outputResult(correctTesting, startTime, testSamples.size());

		if (bestCorrectTesting < correctTesting) {
			bestCorrectTesting = correctTesting;
			bestCorrectTraining = currentCorrect;
			bestNormedMean = normMean;
			bestWindowLength = currentWindowLength;
		}
	}

	double testingCorrect = 1-formatError(bestCorrectTesting, testSamples.size());
	double trainingCorrect = 1-formatError(bestCorrectTraining, trainSamples.size());
	std::cout << trainingCorrect << ";"
			<< testingCorrect << ";"
			<< (bestNormedMean? "true":"false") << ";"
			<< bestWindowLength << std::endl;

}

void printArray(double* data, int size) {
	std::copy(data, data + size, std::ostream_iterator<double>(std::cout, " "));
	std::cout << std::endl;
}

int predictEuclideanDistance(
		const std::vector< TimeSeries* > & testSamples,
		const std::vector< TimeSeries* > & trainSamples) {
	int correct = 0;

	for (auto query : testSamples) {
		double bestSoFar = DOUBLE_MAX_VALUE;
		double label = -1;
		for (auto ts : trainSamples) {
			if (ts != query) {
				double distance = getEuclideanDistance(ts, query, bestSoFar);
				if (distance < bestSoFar) {
					bestSoFar = distance;
					label = ts->getLabel();
				}
			}
		}
		if (query->getLabel() == label) {
			correct++;
		}
	}
	return correct;
}

double getEuclideanDistance (
		TimeSeries* t1,
		TimeSeries* t2,
		double minValue) {
	return getEuclideanDistance(t1->getSize(), t1, t2, 0.0, 1.0, minValue, 0, false);
}

const std::vector<WindowSizes> fit(int maxWindowLength, int minWindowLength, int& accuracy, int& windowSize, bool norm) {
    omp_set_nested( 1 );

	std::vector<WindowSizes> windows;
	windows.resize(maxWindowLength-minWindowLength+1);

	#pragma omp for schedule( dynamic, 1 )
	for (int windowLength = maxWindowLength; windowLength >= minWindowLength; windowLength--) {
		const Predictions p = predict(windowLength, accuracy, norm, trainSamples, trainSamples);

		// determine best window length
		#pragma omp critical (BESTCORRECT)
		{
			windows.push_back({p.correct, windowLength});

			if (p.correct > accuracy
				|| (p.correct == accuracy
					&& windowLength > windowSize)) {
				accuracy = p.correct;
				windowSize = windowLength;
			}

			if (DEBUG) {
				std::cout << p.correct << " "  << windowLength << std::endl;
			}
		}
	}



	return windows;
}


double getEuclideanDistance(
		int windowLength,
		TimeSeries* ts,
		TimeSeries* q,
		double meanTs,
		double stdTs,
		double minValue,
		int w,
		bool normMean) {

	double distance = 0.0;
	double* tsData = ts->getData();
	double* qData = q->getData();
	meanTs = normMean? meanTs : 0;
	for (int ww = 0; ww < windowLength; ww++) {
		double value1 = (tsData[w+ww]-meanTs) * stdTs;
		double value = qData[ww] - value1;
		distance += value*value;

		// early abandoning
		if (distance > minValue) {
			return DOUBLE_MAX_VALUE;
		}
	}
	return distance;
}

const Predictions predictEnsemble(
		std::vector<WindowSizes> windows,
		int bestKnownCorrect,
		bool normMean,
		const std::vector< TimeSeries* > & testSamples,
		const std::vector< TimeSeries* > & trainSamples) {

	omp_set_nested( 1 );

	std::vector<std::vector<double> > testLabels;
	testLabels.resize(testSamples.size());

	// determine the labels for each window
	#pragma omp parallel for
	for (uint i = 0; i < windows.size(); ++i) {
		WindowSizes window = windows[i];
		if ((double)window.accuracy >= ((double)bestKnownCorrect * factor)) { // all with a better score
			Predictions p = predict(window.windowSize, 0, normMean, testSamples, trainSamples);

			#pragma omp critical
			{
				// reorder predictions
				for (uint i = 0; i < testSamples.size(); ++i) {
					testLabels[i].push_back(p.labels[i]);
				}
			}
		}
	}

    int correct = 0;
    std::vector<double> labels;
    labels.resize(testLabels.size());

	// choose the most frequent labels for each sample
    for (uint q = 0; q < testLabels.size(); q++) {
        long maxCount = 0l;
    	double maxLabel = -1;
    	std::tr1::unordered_map<double, long> counts;

    	for (auto label : testLabels[q]) {
			std::tr1::unordered_map<double, long>::iterator iter = counts.find(label);
			int count = 0;
			if (iter == counts.end()) {
				count = 1;
				counts.insert(std::tr1::unordered_map<double, long>::value_type(label, count));
			}
			else {
				count = ++(*iter).second;
			}

			if (count > maxCount
				|| (count == maxCount
					&& maxLabel < label)) {
				maxCount = count;
				maxLabel = label;
			}
		}

    	labels[q] = maxLabel;
    	if (maxLabel == testSamples[q]->getLabel()) {
    		correct++;
    	}
    }

    return {labels, correct};
}

const Predictions predict(
		int windowLength,
		int bestKnownCorrect,
		bool normMean,
		const std::vector< TimeSeries* > & testSamples,
		const std::vector< TimeSeries* > & trainSamples) {

	omp_set_num_threads( 4 );

	volatile int correct = 0;
	volatile int current = 0;

	volatile bool abortLoop = false;

	std::vector<double> labels;
	labels.resize(testSamples.size());

	// calculate means and stds for each sample
	std::vector< std::vector<double> > means(trainSamples.size());
	std::vector< std::vector<double> > stds(trainSamples.size());

	#pragma omp parallel for
	for (uint a = 0; a < trainSamples.size(); ++a) {
		int wLength = min(trainSamples[a]->getSize(), windowLength);
		trainSamples[a]->calcIncreamentalMeanStddev(wLength, means[a], stds[a]);
	}

	#pragma omp parallel
	{
		// iterate each sample to classify
		#pragma omp for schedule( dynamic, 1 )	
		for (uint i = 0; i < testSamples.size(); ++i) {
	   		if (abortLoop) {
				continue;
	   		}

			TimeSeries* query = testSamples[i];
			int wLength = min(query->getSize(), windowLength);
			std::vector<TimeSeries*> disjointWindows = query->getDisjointSequences(wLength, normMean);

			double label = 0.0;
			double distanceTo1NN = DOUBLE_MAX_VALUE;

			for (uint j = 0; (j < trainSamples.size()) && !abortLoop; ++j) {
				TimeSeries* ts = trainSamples[j];

				// Shotgun Distance
				if (ts != query) {

					double totalDistance = 0.0;

					// perform a NN-search
					for (uint w = 0; (w < disjointWindows.size()) && !abortLoop ; ++w) {
						TimeSeries* q = disjointWindows[w];

						// pruning on distance
						if (totalDistance > distanceTo1NN) {
							break;
						}

						// calculate euclidean distances for each sliding window
						// incremental update of windows and distances
						double resultDistance = distanceTo1NN;
						for (int ww = 0, end = ts->getSize()-windowLength+1; (ww < end)  && !abortLoop; ww++) { // faster than reevaluation in for loop
							// calculate the real distance
							double distance = getEuclideanDistance(
									windowLength,
									ts,
									q,
									means[j][ww],
									stds[j][ww],
									resultDistance,
									ww,
									normMean);
							resultDistance = min(distance,resultDistance);
						}
						totalDistance += resultDistance;
					}

					// choose minimum
					if (totalDistance < distanceTo1NN) {
						label = ts->getLabel();
						distanceTo1NN = totalDistance;
					}
				}
			}

			labels[i] = label;

			if (query->getLabel() == label) {
				__sync_fetch_and_add( &correct, 1 );
			}
			__sync_fetch_and_add( &current, 1 );

			for (auto ts : disjointWindows) {
				delete ts;
			}

			// pruning on correct
			if( ( correct+(int)( testSamples.size()-current ) ) < factor*bestKnownCorrect ) {
				abortLoop = true;
			}
		}
	}

	return {labels, correct};
}

double formatError(int correct, int testSize) {
	return (testSize - correct) / (double)testSize;
}

void outputResult(int correct, double startTime, int size) {
	double error = formatError(correct, size);
	std::cout << "The error rate was \t" << std::setprecision(3)
			 << error << "\tcorrect: \t" << (1-error) << "\tTime taken: \t"
			 << (omp_get_wtime() - startTime) << " s" << std::endl;
	std::cout << std::endl;
}

int main(int argc, char* argv[]) {
	std::string path = "classification/";
	for (uint i = 0; i < sizeof(datasets) / sizeof(std::string); i++) {
		trainSamples.clear();
		testSamples.clear();

		std::string dataset = datasets[i];

		std::string trainFilePath = path;
		const char* trainFileName = trainFilePath.append(dataset).append("/").append(dataset).append("_TRAIN").c_str();
		loadDatset(trainFileName, trainSamples);

		std::string testFilePath = path;
		const char* testFileName = testFilePath.append(dataset).append("/").append(dataset).append("_TEST").c_str();
		loadDatset(testFileName, testSamples);

		classify();

		for (std::vector< TimeSeries* >::iterator it = trainSamples.begin(); it != trainSamples.end(); ++it) {
			delete *it;
		}
		for (std::vector< TimeSeries* >::iterator it = testSamples.begin(); it != testSamples.end(); ++it) {
			delete *it;
		}

	}

}
